#!/usr/bin/env python
import numpy as np


def saltelli(params, nbase, lhs=False, nskip=1):
    """
        Samples parameters in given ranges so they can be used
        for the calculation of Sobol indices as described by
        Saltelli (2002) and Saltelli et al. (2007).


        Definition
        ----------
        def saltelli(params, nbase, lhs=False, nskip=1):


        Input
        -----
        params    (nparams,2) array of parameter ranges
        nbase     base sample size


        Optional Input
        --------------
        lhs       if True: use latin hypercube sampling in uniform distributions
                  if False: use sobol sequence
                            This uses the package sobol of Corrado Chisari
        nskip     number of sobol sequences to skip at the beginning


        Output
        ------
        array with (nparams,nbase*(nparam+2)) with sample values in the right order.


        Restrictions
        ------------
        1. Only uniform latin hypercube sampling.
        2. No parameter dependencies possible.


        References
        ----------
        Saltelli, A. (2002). Making best use of model evaluations to compute sensitivity indices.
            Computer Physics Communications, 145(2), 280-297.
        Saltelli, A. et al. (2008). Global sensitivity analysis. The primer.
            John Wiley & Sons Inc., NJ, USA, ISBN 978-0-470-05997-5 (pp. 1-292)


        Examples
        --------
        >>> import numpy as np
        >>> params = np.array([[  1.00000000e+00,   1.00000000e+06],
        ...                    [  1.00000000e+00,   2.00000000e+01]])
        >>> nbase = 10
        >>> # seed for reproducible results in doctest
        >>> np.random.seed(1)
        >>> out = saltelli(params, nbase)
        >>> from autostring import astr
        >>> print(astr(out[0:2,0:4],3,pp=True))
        [['     1.000' '500000.500' '750000.250' '250000.750']
         ['     1.000' '    10.500' '     5.750' '    15.250']]

        >>> out = saltelli(params, nbase, nskip=2)
        >>> print(astr(out[0:2,0:4],3,pp=True))
        [['500000.500' '750000.250' '250000.750' '375000.625']
         ['    10.500' '     5.750' '    15.250' '     8.125']]

        >>> out = saltelli(params, nbase, nskip=2, lhs=True)
        >>> print(astr(out[0:2,0:4],3,pp=True))
        [['341702.859' '872032.577' '500011.937' '930233.327']
         ['     1.796' '     6.102' '    16.588' '     4.568']]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2012
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Apr 2014 - assert
    """
    #
    # Check input
    assert np.size(params[0,:]) == 2, 'parameter ranges must be given in form (nparams,2).'
    nparams = np.size(params[:,0])
    #
    nso = nbase*(nparams+2)
    # A=0:nbase, B=nbase:nbase*2, C=nbase*2:nso
    zoff = params[:,0]
    zmul = params[:,1] - params[:,0]
    # Two random samples A and B
    pA   = np.empty((nparams,nbase))
    pB   = np.empty((nparams,nbase))
    if lhs:
        import scipy.stats as stats
        from pyjams.jams.lhs import lhs
        dist = [stats.uniform     for i in range(nparams)]
        pars = [(zoff[i],zmul[i]) for i in range(nparams)]
        dist = dist + dist # 2*nparams
        pars = pars + pars
        lat  = lhs(dist, pars, nbase)
        for i in range(nparams):
            pA[i,:] = lat[i,:]
            pB[i,:] = lat[i+nparams,:]
    else:
        import sobol
        sob = sobol.i4_sobol_generate(2*nparams,nbase,nskip)
        for i in range(nparams):
            pA[i,:] = zoff[i] + zmul[i]*sob[i,:]
            pB[i,:] = zoff[i] + zmul[i]*sob[i+nparams,:]
    # The C sample is nparams the B sammple
    pC = np.array([pB for i in range(nparams)])
    # where on each repeat one column is replaced by the column of A
    for i in range(nparams):
        pC[i,i,:] = pA[i,:]

    # Reshape so that one can do runs over 2nd dim and then use jams.sobol_index
    pout = np.empty((nparams,nso))
    for i in range(nparams):
        pout[i,:] = np.concatenate((pA[i,:],pB[i,:],np.ravel(pC[:,i,:])))

    return pout


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    # params = np.array([[  1.00000000e+00,   1.00000000e+06],
    #                    [  1.00000000e+00,   2.00000000e+01]])
    # nbase = 10
    # out = saltelli(params, nbase)
    # print out
    # out = saltelli(params, nbase, nskip=2)
    # print out
    # out = saltelli(params, nbase, nskip=2,lhs=True)
    # print out




#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import smtplib

__all__ = ['sendmail']

def sendmail(fromm, to, subject, message, login, password, cc=None, smtpserver='smtp.ufz.de:587'):
    """
        Send an e-mail.

        The default smtp-server is smtp.ufz.de on port 587 (SSL/START-TLS).


        Definition
        ----------
        def sendmail(fromm, to, subject, message, login, password, cc=None, smtpserver='smtp.ufz.de:587'):


        Input
        -----
        fromm        string, Name or e-mail for From field.
        to           string, Comma separated list of e-mail addresses.
        subject      string, Subject line
        message      string, Message text
        login        login at smtp-server
        password     password for smtp-server


        Optional input parameters
        -------------------------
        cc           string, Comma separated list of e-mail addresses.
        smtpserver   string, SMTP-server:port (default: smtp.ufz.de:587)


        Output
        ------
        Return of smtplib sendmail function


        Examples
        --------
        import getpass
        email = input('Enter e-mail address in (double) quotes (\' \' or " "): ')
        user = getpass.getuser()
        prompt = 'Enter password for user '+user+': '
        ipass = getpass.getpass(prompt)

        a = sendmail('sendmail@ufz.de',
                     to       = email,
                     subject  = 'sendmail test',
                     message  = '''
                                This is the actual message.
                                It can have several lines, etc.

                                Regards
                                Matthias
                                ''',
                    login    = user,
                    password = ipass)
        print('Sendmail return: ', a ,'.')


        Notes
        -----
        modified from http://rosettacode.org/wiki/Send_an_email#Python


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """

# --------------------------------------------------------------------
# Send E-Mail
# modified from http://rosettacode.org/wiki/Send_an_email#Python
    header  = 'From: {:s}\n'.format(fromm)
    if type(to) == list:
        too     = to                                 # to field in sendmail needs list
        header += 'To: {:s}\n'.format(','.join(to))  # To: in message needs string
    else:
        too     = to.split(',')
        header += 'To: {:s}\n'.format(to)
    if (cc is not None) and (cc != ''):
        if type(cc) == list:
            header += 'Cc: {:s}\n'.format(','.join(cc))
        else:
            header += 'Cc: {:s}\n'.format(cc)
    header += 'Subject: {:s}\n\n'.format(subject)
    message = header + message

    server = smtplib.SMTP(smtpserver)
    server.starttls()
    server.login(login, password)
    problems = server.sendmail(fromm, too, message)
    server.quit()
    return problems


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

#     import getpass
#     email = input('Enter e-mail address in (double) quotes (\' \' or " "): ')
#     user = getpass.getuser()
#     prompt = 'Enter password for user '+user+': '
#     ipass = getpass.getpass(prompt)

#     a = sendmail('sendmail@ufz.de',
#                  to       = email,
#                  subject  = 'sendmail test',
#                  message  = '''
# This is the actual message.
# It can have several lines, etc.

# Regards
# Matthias
#                             ''',
#                 login    = user,
#                 password = ipass)
#     print('Sendmail return: ', a ,'.')



#!/usr/bin/env python
"""
Median absolute deviation test

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2011-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   mad

History
    * Written Nov 2011 by Matthias Cuntz - mc (at) macu (dot) de
    * ND-array, act on axis=0, May 2012, Matthias Cuntz
    * Removed bug in broadcasting, axis=0 did not always work: spread md and
      MAD to input dimensions, Jun 2012, Matthias Cuntz
    * Better usage of numpy possibilities, e.g. using np.diff,
      remove spreads, Jun 2012, Matthias Cuntz
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Use bottleneck for medians, otherwise loop over axis=1,
      Jul 2013, Matthias Cuntz and Juliane Mai
    * Re-allow masked arrays and arrays with NaNs, Jul 2013, Matthias Cuntz
    * Removed bug in NaN treatment, Oct 2013, Matthias Cuntz
    * Keyword nonzero, Oct 2013, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * Ported to pyjams, Jan 2022, Matthias Cuntz
    * Only one test of bottleneck availability, Jan 2022, Matthias Cuntz
    * Return all False instead of all True if all masked before MAD started,
      Jan 2022, Matthias Cuntz
    * prepend, append as in numpy.diff, May 2023, Matthias Cuntz
    * Support pandas Series and DataFrame, Jul 2023, Matthias Cuntz

"""
import numpy as np
from .helper import input2array, array2input


__all__ = ['mad']


def mad(datin, z=7, deriv=0, nozero=False, prepend=None, append=None):
    """
    Median absolute deviation test

    The test acts either on raw values, or on 1st or 2nd derivatives.

    Return mask with True where value is out of range, i.e.
    :math:`< (median - z.MAD/0.6745)` or :math:`> (md + z.MAD/0.6745)`.

    Parameters
    ----------
    datin : array or masked array
        `mad` acts on ``axis=0``
    z : float, optional
        Input is allowed to deviate maximum *z* (estimators of) standard
        deviations from the median (default: 7)
    deriv : int, optional
        Act on raw input (0, default), on first derivatives (1), or on 2nd
        derivatives (2)
    nozero : bool, optional
        Exclude zeros (0.) from input *datin* if True.
    prepend, append : array_like, optional
        Values to prepend or append to `datin` prior to performing the
        difference with `numpy.diff` if `deriv > 0`.
        `prepend` uses `numpy.insert` and `append` uses `numpy.append`.
        Scalar values are hence expanded to arrays with length 1 in the first
        axis and the shape of the input array along all other axes. Otherwise
        the dimension and shape must match `datin` except along the first axis.

        .. versionadded:: 1.31

    Returns
    -------
    array of bool
        False everywhere except where input deviates more than
        *z* standard deviations from median.
        The shape of the output is the same as `datin` except for the first
        dimension, which is smaller by `deriv` if `prepend` and `append`
        are not set.

    Notes
    -----
    If input is an ndarray then mad is checked along the first axis for
    outliers.

    The 1st derivative is calculated simply as
    ``d = numpy.diff(datin, n=1, axis=0)`` because mean of left and right
    difference would give 0 for spikes.

    The 2nd derivative is calculated as ``d = numpy.diff(datin, n=2, axis=0)``.

    If ``numpy.all(d.mask)`` then ``d.mask`` is returned, which is all True.

    NaN does not return True because this would remove points adjacent to NaN
    if ``deriv > 0``.

    Examples
    --------
    >>> import numpy as np
    >>> y = np.array([-0.25, 0.68, 0.94, 1.15, 2.26, 2.35, 2.37, 2.40, 2.47,
    ...               2.54, 2.62, 2.64, 2.90, 2.92, 2.92, 2.93, 3.21, 3.26,
    ...               3.30, 3.59, 3.68, 4.30, 4.64, 5.34, 5.42, 8.01])

    MAD on raw data

    >>> print(mad(y))
    [False False False False False False False False False False False False
     False False False False False False False False False False False False
     False False]
    >>> print(mad(y, z=4))
    [False False False False False False False False False False False False
     False False False False False False False False False False False False
     False  True]
    >>> print(mad(y, z=3))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]

    MAD on 2nd derivatives

    >>> print(mad(y, z=4, deriv=2))
    [False False False False False False False False False False False False
     False False False False False False False False False False False  True]

    MAD on 2nd derivatives with prepend and append set

    >>> print(mad(y, z=4, deriv=2, prepend=y[0], append=y[-1]))
    [False False False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]

    Use for masking arrays, for example

    >>> my = np.ma.array(y, mask=mad(y, z=4))
    >>> print(my)
    [-0.25 0.68 0.94 1.15 2.26 2.35 2.37 2.4 2.47 2.54 2.62 2.64 2.9 2.92 2.92
     2.93 3.21 3.26 3.3 3.59 3.68 4.3 4.64 5.34 5.42 --]

    MAD on several dimensions

    >>> yy = np.transpose(np.array([y, y]))
    >>> print(np.transpose(mad(yy, z=4)))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False False
      False  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False False
      False  True]]
    >>> yyy = np.transpose(np.array([y, y, y]))
    >>> print(np.transpose(mad(yyy, z=3)))
    [[ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [ True False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]]
    >>> print(np.transpose(mad(yy, z=4, deriv=2)))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False  True]]

    Set prepend and append either as scalar or array

    >>> print(np.transpose(mad(yy, z=4, deriv=2, prepend=y[0], append=y[-1])))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]]
    >>> print(np.transpose(mad(yy, z=4, deriv=2,
    ...                        prepend=yy[0, :], append=yy[-1, :])))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]]
    >>> print(np.transpose(mad(yy, z=4, deriv=2,
    ...                        prepend=yy[0:1, :], append=yy[-1:, :])))
    [[False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]
     [False False False False False False False False False False False False
      False False False False False False False False False False False False
       True  True]]

    Masked arrays

    >>> my = np.ma.array(y, mask=np.zeros(y.shape))
    >>> my.mask[-1] = True
    >>> print(mad(my, z=4))
    [True False False False False False False False False False False False
     False False False False False False False False False False False False
     False --]
    >>> print(mad(my, z=3))
    [True False False False False False False False False False False False
     False False False False False False False False False False False True
     True --]

    Arrays with NaNs

    >>> ny = y.copy()
    >>> ny[-1] = np.nan
    >>> print(mad(ny, z=4))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
     False False]
    >>> print(mad(ny, z=3))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False  True
      True False]

    Exclude zeros

    >>> zy = y.copy()
    >>> zy[1] = 0.
    >>> print(mad(zy, z=3))
    [ True  True False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]
    >>> print(mad(zy, z=3, nozero=True))
    [ True False False False False False False False False False False False
     False False False False False False False False False False False False
      True  True]

    """
    if isinstance(datin, np.ma.MaskedArray):
        idatin = datin.copy()
    else:
        idatin = input2array(datin)

    assert (deriv >= 0) and (deriv <= 2), (f'deriv > 2 unimplemented: {deriv}')
    if deriv == 1:
        if (prepend is not None) and (append is not None):
            raise ValueError('Only one of prepend and append may be given if'
                             ' deriv==1.')

    # prepend and append
    if deriv > 0:
        if prepend is not None:
            idatin = np.insert(idatin, 0, prepend, axis=0)

        if append is not None:
            if np.iterable(append):
                iappend = np.array(append)
                # all but first dimension must match
                shpdat = np.array(idatin.shape)
                shpappend = np.array(iappend.shape)
                if iappend.ndim == idatin.ndim:
                    assert np.all(shpdat[1:] == shpappend[1:]), (
                        f'Shape of append {iappend.shape} must match shape'
                        f' of datin {idatin.shape} except first dimension.')
                elif iappend.ndim == (idatin.ndim - 1):
                    assert np.all(shpdat[1:] == shpappend), (
                        f'Shape of append {iappend.shape} must match shape'
                        f' of datin {idatin.shape} without first dimension.')
                    iappend = iappend[np.newaxis, ...]
                else:
                    raise ValueError(
                        f'Shape of append {iappend.shape} must match shape'
                        f' of datin {idatin.shape} except or without first'
                        f' dimension.')
            else:
                shpdat = list(idatin.shape)
                shpdat[0] = 1
                iappend = np.full(shpdat, append)
            idatin = np.append(idatin, iappend, axis=0)

    if nozero:
        ii = np.where(idatin == 0.)[0]
        if ii.size > 0:
            idatin[ii] = np.nan

    # make derivative
    sn = list(np.shape(idatin))
    n  = sn[0]
    if deriv == 0:
        m      = n
        d      = idatin
    elif deriv == 1:
        m      = n - 1
        sm     = sn
        sm[0]  = m
        d      = np.diff(idatin, axis=0)
    elif deriv == 2:
        m      = n - 2
        sm     = sn
        sm[0]  = m
        d      = np.diff(idatin, n=2, axis=0)
    else:  # pragma: no cover
        raise ValueError('Should not be here')

    # Shortcut if all masked
    ismasked = isinstance(d, np.ma.MaskedArray)
    if not ismasked:
        ii = np.where(~np.isfinite(d))[0]
        d  = np.ma.array(d)
        if ii.size > 0:
            d[ii] = np.ma.masked

    if np.all(d.mask):
        if ismasked:
            return d.mask
        else:
            res = np.zeros(d.shape, dtype=bool)
            res = array2input(res, datin)
            return res

    # Use bottleneck if available
    try:     # pragma: no cover
        import bottleneck as bn
        med = bn.median
    except:  # pragma: no cover
        med = np.median

    # Median
    oldsettings = np.geterr()
    np.seterr(invalid='ignore')
    if d.ndim == 1:
        dd = d.compressed()
        md = med(dd)
        # Median absolute deviation
        MAD = med(np.abs(dd - md))
        # Range around median
        thresh = MAD * (z / 0.6745)
        # True where outside z-range
        res = (d < (md - thresh)) | (d > (md + thresh))
    elif d.ndim == 2:
        res = np.empty(d.shape, dtype=bool)
        for i in range(d.shape[1]):
            di = d[:, i]
            dd = di.compressed()
            md = med(dd)
            # Median absolute deviation
            MAD = med(np.abs(dd - md))
            # Range around median
            thresh = MAD * (z / 0.6745)
            # True where outside z-range
            res[:, i] = (d[:, i] < (md - thresh)) | (d[:, i] > (md + thresh))
    else:
        np.seterr(**oldsettings)
        raise ValueError('idatin.ndim must be <= 2')

    np.seterr(**oldsettings)
    if ismasked:
        return res
    else:
        # got masked because of NaNs
        if isinstance(res, np.ma.MaskedArray):
            res = np.where(res.mask, False, res)
        res = array2input(res, datin)
        return res


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from pyjams.const import eps


def zacharias(h, clay, sand, db, params=None, thetar=False, thetas=False, lnalpha=False, n=False):
    """
        Soil water content with the van Genuchten equation and
        the pedotransfer functions of Zacharias et al. (2007).


        Definition
        ----------
        def zacharias(h, clay, sand, db, params=None, thetar=False, thetas=False, lnalpha=False, n=False):


        Input
        -----
        h          pressure head, scalar or array [cm], 0=saturation, 15000=wilting point
        clay       clay content, scalar or array [%, i.e. 0-100]
        sand       sand content, scalar or array [%, i.e. 0-100]
        db         bulk density, scalar or array [g/cm3], quartz=2.65


        Optional Input
        --------------
        params     Parameter for Zacharias et al. (2007) pedotransfer functions
                   If None, values from Zacharias et al. will be taken that are different
                   between sandy and non-sandy soil (<66.5% sand)


        Options
        -------
        thetar     If True, outputs residual water content thetar as well [m3 m-3]
        thetas     If True, outputs saturation water content thetas as well [m3 m-3]
        lnalpha    If True, outpus logarithm of shape parameter alpha as well [1/cm]
        n          If True, output exponent n as well


        Output
        ------
        Soil water content theta [m^3 m^-3]


        Restrictions
        ------------
        Does not check the validity of the parameter set, i.e. negative soil moistures
        can occur, for example.
        Use zacharias_check to check the parameter set first.


        Examples
        --------
        >>> h = np.array([0.0000000, 0.0000000, 10.000000, 31.622777,
        ...               100.00000, 199.52623, 199.52623,
        ...               501.18723, 2511.8864, 15848.932])
        >>> sand = np.array([12.800000, 61.600000, 17.200000, 85.800000,
        ...                  16.500000, 12.800000, 61.600000,
        ...                  17.200000, 85.800000, 16.500000])
        >>> clay = np.array([30.500000, 17.200000, 25.500000, 8.9000000,
        ...                  28.100000, 30.500000, 17.200000,
        ...                  25.500000, 8.9000000, 28.100000])
        >>> rho = np.array([1.2100000, 1.3400000, 1.4600000, 1.6300000,
        ...                 1.3000000, 1.2100000, 1.3400000,
        ...                 1.4600000, 1.6300000, 1.3000000])
        >>> from autostring import astr
        >>> print(astr(zacharias(h, clay, sand, rho),3,pp=True))
        ['0.500' '0.453' '0.421' '0.245' '0.393' '0.381' '0.285' '0.313' '0.039' '0.221']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jun 2012
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Nov 2016 - const.tiny -> const.eps
    """
    #
    # Check input
    ih = np.where(h==0., eps, h)
    if np.any(ih < 0.) | np.any(ih > 1e6):
        raise ValueError('h must be >=0 and <= 1e6 (=pf6)')
    iclay = np.where(clay==0., eps, clay)
    if np.any(iclay < 0.) | np.any(iclay > 100.):
        raise ValueError('clay must be >=0 and <= 100.')
    isand = np.where(sand==0., eps, sand)
    if np.any(isand < 0.) | np.any(isand > 100.):
        raise ValueError('sand must be >=0 and <= 100.')
    idb = np.array(db)
    if np.any(idb < 0.) | np.any(idb > 2.65):
        raise ValueError('db must be >=0 and <= 2.65.')
    nn = np.size(isand)
    if (np.size(iclay) != nn) | (np.size(idb) != nn) | (np.size(ih) != nn):
        raise ValueError('h, sand, clay, and db must have the same sizes.')
    if params is not None:
        if np.size(params) != 15:
            raise ValueError('size(params) must be 15.')
    # save output shape
    ns    = np.shape(isand)
    iclay = np.ravel(iclay)
    isand = np.ravel(isand)
    idb   = np.ravel(idb)
    # Take right params
    par0  = np.empty(nn)
    par1  = np.empty(nn)
    par2  = np.empty(nn)
    par3  = np.empty(nn)
    par4  = np.empty(nn)
    par5  = np.empty(nn)
    par6  = np.empty(nn)
    par7  = np.empty(nn)
    par8  = np.empty(nn)
    par9  = np.empty(nn)
    par10 = np.empty(nn)
    par11 = np.empty(nn)
    par12 = np.empty(nn)
    par13 = np.empty(nn)
    par14 = np.empty(nn)
    if params is not None:
        # Either params given
        par0[:]  = params[0]
        par1[:]  = params[1]
        par2[:]  = params[2]
        par3[:]  = params[3]
        par4[:]  = params[4]
        par5[:]  = params[5]
        par6[:]  = params[6]
        par7[:]  = params[7]
        par8[:]  = params[8]
        par9[:]  = params[9]
        par10[:] = params[10]
        par11[:] = params[11]
        par12[:] = params[12]
        par13[:] = params[13]
        par14[:] = params[14]
    else:
        # or take Zacharias
        parclay = np.array([ 0.,     0.,     0.,
                             0.788,  0.001, -0.263,
                            -0.648,  0.044, -3.168,  0.023,
                             1.392,  1.212, -0.704, -0.418, -0.024])
        parsand = np.array([ 0.,     0.,     0.,
                             0.890, -0.001, -0.322,
                            -4.197,  0.076, -0.276, 0.013,
                            -2.562,  3.750, -0.016, 7e-9,  4.004])
        for i in range(nn):
            if isand[i] < 66.5:
                par = parclay
            else:
                par = parsand
            par0[i]  = par[0]
            par1[i]  = par[1]
            par2[i]  = par[2]
            par3[i]  = par[3]
            par4[i]  = par[4]
            par5[i]  = par[5]
            par6[i]  = par[6]
            par7[i]  = par[7]
            par8[i]  = par[8]
            par9[i]  = par[9]
            par10[i] = par[10]
            par11[i] = par[11]
            par12[i] = par[12]
            par13[i] = par[13]
            par14[i] = par[14]

    # Zacharias pedotransfer
    ithetar = par0  + par1*iclay                        + par2*idb
    ithetas = par3  + par4*iclay                        + par5*idb
    ilna    = par6  + par7*iclay                        + par8*idb + par9*isand
    inn     = par10 + par11*np.exp(par12*np.log(iclay))            + par13*np.exp(par14*np.log(isand))
    imm     = 1. - 1./np.where(inn != 0., inn, 1e-3)
    # van Genuchten sign
    # limit exp to 600 and log to eps so that no over- and underflows occur
    expmax  = 600.
    lnah    = np.log(np.maximum(np.exp(ilna)*ih, eps))
    ahn     = np.exp(np.minimum(inn*lnah, expmax))
    denom   = np.maximum(np.exp(np.minimum(imm*np.log(1.+ahn), expmax)), eps)
    itheta  = np.where(ih <= eps, ithetas, ithetar + (ithetas-ithetar)/denom)
    # Output
    itheta = np.reshape(itheta, ns)
    if nn==1: itheta = np.float(itheta)
    if (thetar==False) & (thetas==False) & (lnalpha==False) & (n==False):
        return itheta
    else:
        out = [itheta]
        if thetar==True:
            ithetar = np.reshape(ithetar, ns)
            if nn==1: ithetar = np.float(ithetar)
            out = out + [ithetar]
        if thetas==True:
            if nn==1: ithetas = np.float(ithetas)
            ithetas = np.reshape(ithetas, ns)
            out = out + [ithetas]
        if lnalpha==True:
            if nn==1: ilna = np.float(ilna)
            ilna = np.reshape(ilna, ns)
            out = out + [ilna]
        if n==True:
            if nn==1: inn = np.float(inn)
            inn = np.reshape(inn, ns)
            out = out + [inn]
        return out

def zacharias_check(params, sand=None, clay=None):
    """
        Checks if a given parameter set is valid for all possible soils with
        the van Genuchten equation and the pedotransfer functions of Zacharias et al. (2007).

        Definition
        ----------
        def zacharias_check(params, sand=None, clay=None):


        Input
        -----
        params     array[15] with parameters a1, b1, ..., d4 of Zacharias et al. (2007)


        Optional Input
        --------------
        clay       If given: 1. < clay < 99. then calc sand content < clay
        sand       If given: 1. < sand < 99. then calc sand content > sand


        Output
        ------
        Boolean    True:  valid parameter set for all soils
                   False: not valid in at least one extreme case


        Examples
        --------
        >>> parclay = np.array([ 0.,     0.,     0.,
        ...                      0.788,  0.001, -0.263,
        ...                     -0.648,  0.044, -3.168,  0.023,
        ...                      1.392,  1.212, -0.704, -0.418, -0.024])
        >>> print(zacharias_check(parclay))
        True
        >>> print(zacharias_check(parclay, clay=66))
        True
        >>> parsand = np.array([ 0.,     0.,     0.,
        ...                      0.890, -0.001, -0.322,
        ...                     -4.197,  0.076, -0.276, 0.013,
        ...                     -2.562,  3.750, -0.016, 7e-9,  4.004])
        >>> print(zacharias_check(parsand))
        False
        >>> print(zacharias_check(parsand, sand=66))
        True


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2012
        Modified, MC, Feb 2013 - ported to Python 3
    """
    #
    # Check input
    if params is not None:
        if np.size(params) != 15: raise ValueError('size(params) must be 15.')
    # Check ranges
    isgood = True
    if (params[0]  <   0.)   |  (params[0]  >  1.):   isgood = False
    if (params[1]  <  -0.01) |  (params[1]  >  0.01): isgood = False
    if (params[2]  <  -0.5)  |  (params[2]  >  0.5):  isgood = False
    if (params[3]  <   0.)   |  (params[3]  >  1.):   isgood = False
    if (params[4]  <  -0.01) |  (params[4]  >  0.01): isgood = False
    if (params[5]  <  -0.5)  |  (params[5]  >  0.5):  isgood = False
    if (params[6]  < -10.)   |  (params[6]  > 10.):   isgood = False
    if (params[7]  <  -1.)   |  (params[7]  >  1.):   isgood = False
    if (params[8]  <  -5.)   |  (params[8]  >  5.):   isgood = False
    if (params[9]  <  -1.)   |  (params[9]  >  1.):   isgood = False
    if (params[10] < -10.)   |  (params[10] > 10.):   isgood = False
    if (params[11] < -10.)   |  (params[11] > 10.):   isgood = False
    if (params[12] <  -5.)   |  (params[12] >  5.):   isgood = False
    if (params[13] < -10.)   |  (params[13] > 10.):   isgood = False
    if (params[14] <  -5.)   |  (params[14] >  5.):   isgood = False
    # Soil ranges
    h    = [1., 1e6]
    if clay is not None:
        rclay = [98., 1.,   99.-clay]
        rsand = [1.,  clay, clay]
    elif sand is not None:
        rclay = [1.,  1.,   99.-sand]
        rsand = [98., sand, sand]
    else:
        rclay = [1.,  1., 98.]
        rsand = [98., 1.,  1.]
    db   = [0.5, 2.3]
    nn   = 2*3*2
    ih    = np.empty(nn)
    iclay = np.empty(nn)
    isand = np.empty(nn)
    idb   = np.empty(nn)
    zaehl = 0
    for i in [0,1]:
        for j in [0,1,2]:
            for k in [0,1]:
                ih[zaehl]    = h[i]
                iclay[zaehl] = rclay[j]
                isand[zaehl] = rsand[j]
                idb[zaehl]   = db[k]
                zaehl += 1
    # Calc Zacharias
    itheta, ithetar, ithetas, ilna, inn = zacharias(ih, iclay, isand, idb, params=params,
                                                    thetar=True, thetas=True, lnalpha=True, n=True)
    #ia  = np.exp(ilna)
    #imm = 1. - 1./inn
    if np.any(ithetar < 0.) | np.any(ithetar > 1.): isgood = False
    if np.any(ithetas < 0.) | np.any(ithetas > 1.) | np.any(ithetas < ithetar): isgood = False
    #if np.any(ilna > 0.): isgood = False
    if np.any(inn < 1.) | np.any(inn > 10.): isgood = False
    #if np.any(imm < 0.): isgood = False
    #if np.any(ia > 1.): isgood = False
    if np.any(itheta < ithetar) | np.any(itheta > ithetas): isgood = False

    return isgood


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # parclay = np.array([ 0.,     0.,     0.,
    #                      0.788,  0.001, -0.263,
    #                      -0.648,  0.044, -3.168,  0.023,
    #                      1.392,  1.212, -0.704, -0.418, -0.024])
    # print zacharias_check(parclay)
    # print zacharias_check(parclay, clay=66)
    # parsand = np.array([ 0.,     0.,     0.,
    #                      0.890, -0.001, -0.322,
    #                      -4.197,  0.076, -0.276, 0.013,
    #                      -2.562,  3.750, -0.016, 7e-9,  4.004])
    # print zacharias_check(parsand)
    # print zacharias_check(parsand, sand=66)

    # h = np.array([0.0000000, 0.0000000, 10.000000, 31.622777,
    #               100.00000, 199.52623, 199.52623,
    #               501.18723, 2511.8864, 15848.932])
    # sand = np.array([12.800000, 61.600000, 17.200000, 85.800000,
    #                  16.500000, 12.800000, 61.600000,
    #                  17.200000, 85.800000, 16.500000])
    # clay = np.array([30.500000, 17.200000, 25.500000, 8.9000000,
    #                  28.100000, 30.500000, 17.200000,
    #                  25.500000, 8.9000000, 28.100000])
    # rho = np.array([1.2100000, 1.3400000, 1.4600000, 1.6300000,
    #                 1.3000000, 1.2100000, 1.3400000,
    #                 1.4600000, 1.6300000, 1.3000000])
    # # # From IDL code
    # # zthetafit = np.array([0.50026998, 0.45277997, 0.42107488, 0.24487569,
    # #                       0.39254300, 0.38144154, 0.28500143,
    # #                       0.31320843, 0.038888339, 0.22115273])
    # print zacharias(h, clay, sand, rho)
    # # [ 0.50027     0.45278     0.42107491  0.24487575  0.39254299  0.38144151
    # #   0.28500141  0.31320842  0.03888839  0.22115267]




#!/usr/bin/env python
"""
Test functions for parameter sensitivity analysis.

They were taken from

    Ishigami and Homma (1990)
        An importance qualification technique in uncertainty analysis for
        computer models, Proceedings of the isuma '90, First International
        Symposium on Uncertainty Modelling and Analysis, University of
        Maryland, Dec. 03 - Dec 05 1990, 398-403

    Oakley and O'Hagan (2004)
        Probabilistic sensitivity analysis of complex models: a Bayesian
        approach J. R. Statist. Soc. B 66, Part 3, 751-769.

    Morris (1991)
        Factorial sampling plans for preliminary computational experiments,
        Technometrics 33, 161-174.

    Saltelli et al. (2008)
        Global Sensitivity Analysis. The Primer, John Wiley & Sons, pp. 292

    Saltelli et al. (2010)
        Variance based sensitivity analysis of model output, Design and
        estimator for the total sensitivity index, Comp. Phys. Comm. 181,
        259-270.

    Sobol' (1990)
        Sensitivity estimates for nonlinear mathematical models,
        Matematicheskoe Modelirovanie 2, 112-118 (in Russian),
        translated in English in Sobol' (1993).

    Sobol' (1993)
        Sensitivity analysis for non-linear mathematical models,
        Mathematical Modelling and Computational Experiment 1, 407-414,
        English translation of Russian original paper Sobol' (1990).

Current functions are:

.. list-table::
   :widths: 15 50
   :header-rows: 1

   * - Function
     - Description
   * - B
     - B of Saltelli et al. (2010)
   * - G / g
     - G-function attributed to Sobol' (1990, 1993), given by Saltelli et al.
       (2008, 2010)
   * - Gstar
     - G* of Saltelli et al. (2010)
   * - ishigami_homma
     - Ishigami and Homma (1990), given by Saltelli et al. (2008, page 179)
   * - K / bratley
     - K of Saltelli et al. (2010)
   * - fmorris / morris
     - After Morris (1991)
   * - oakley_ohagan
     - Oakley and O'Hagan (2004), parameters given in Saltelli et al. (2008) or
       on http://www.jeremy-oakley.staff.shef.ac.uk/psa_example.txt
   * - linear
     - Linear test function :math:`Y = a*X + b`
   * - product
     - Product test function :math:`Y = X[0] * X[1]`
   * - ratio
     - Ration test function :math:`Y = X[0] / X[1]`
   * - ishigami_homma_easy
     - Simplified Ishigami and Homma function :math:`Y = sin(X[0]) + X[1]`

This module was written by Matthias Cuntz & Juliane Mai while at
Department of Computational Hydrosystems, Helmholtz Centre for
Environmental Research - UFZ, Leipzig, Germany, and continued by
Matthias Cuntz while at Institut National de Recherche pour l'Agriculture,
l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2015-2022 Matthias Cuntz, Juliane Mai, see AUTHORS.rst
            for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

Functions:

.. autosummary::
   B
   g
   G
   Gstar
   K
   bratley
   fmorris
   morris
   oakley_ohagan
   ishigami_homma
   linear
   product
   ratio
   ishigami_homma_easy

History
    * Written Mar 2015 by Matthias Cuntz (mc (at) macu (dot) de) & Juliane Mai
    * Added functions to test PAWN method properly: linear, product, ratio, and
      ishigami_homma_easy, Dec 2017, Juliane Mai
    * Provide morris function under the name fmorris and the K function under
      the name bratley, Nov 2019, Matthias Cuntz
    * Changed to Sphinx docstring and numpydoc, Dec 2019, Matthias Cuntz
    * Distinguish iterable and array_like parameter types,
      Jan 2020, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""
import numpy as np


__all__ = ['B', 'g', 'G', 'Gstar', 'K', 'bratley', 'fmorris', 'morris',
           'oakley_ohagan', 'ishigami_homma',
           'linear', 'product', 'ratio', 'ishigami_homma_easy']


# -----------------------------------------------------------

def B(X):
    """
    B function

    Saltelli et al. (2010) Comp. Phys. Comm. 181, p. 259-270

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats

    Returns
    -------
    B : float or ndarray
        float or (npoints,) floats of B function values at *X*

    """
    # Parameter sets are assumed to be in following ordering:
    # (x_1, x_2, ..., X_m, w_1, w_2, ..., w_m)
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] % 2 == 0, 'X.shape[0] must be even.'

    m  = iX.shape[0] // 2
    y  = np.sum(iX[:m, :] * iX[m:, :], axis=0)

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def g(X, a):
    """
    G-function

    Sobol' (1990) Matematicheskoe Modelirovanie 2, 112-118 (in Russian)

    Sobol' (1993) Mathematical Modelling and Computational Experiment 1,
    407-414 (English translation)

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats
    a : array_like
        (nX,) array of floats

    Returns
    -------
    G : float or ndarray
        float or (npoints,) floats of G function values at *X*
        with parameters *a*

    """
    return Gstar(X, np.ones(len(a)), np.zeros(len(a)), a)


# -----------------------------------------------------------


def G(X, a):
    """
    G-function

    Sobol' (1990) Matematicheskoe Modelirovanie 2, 112-118 (in Russian)

    Sobol' (1993) Mathematical Modelling and Computational Experiment 1,
    407-414 (English translation)

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats
    a : array_like
        (nX,) array of floats

    Returns
    -------
    g : float or ndarray
        float or (npoints,) floats of G function values at *X*
        with parameters *a*

    """
    return Gstar(X, np.ones(len(a)), np.zeros(len(a)), a)


# -----------------------------------------------------------


def Gstar(X, alpha, delta, a):
    """
    G* example

    Saltelli et al. (2010) Comp. Phys. Comm., 181, p. 259-270

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats
    alpha : array_like
        (nX,) array of floats
    delta : array_like
        (nX,) array of floats
    a : array_like
        (nX,) array of floats

    Returns
    -------
    G* : float or ndarray
        float or (npoints,) floats of G* function values at *X* with
        parameters *alpha*, *delta* and *a*

    """
    # Model output for given parameter set(s) is returned
    # X: dim1 = # of parameters = 10
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    if isinstance(alpha, (list, tuple)):
        nalpha = len(alpha)
    else:
        nalpha = alpha.size

    assert iX.shape[0] == nalpha, 'X.shape[0] must len(alpha).'

    ialpha = np.array(alpha).reshape((nalpha, 1))
    idelta = np.array(delta).reshape((nalpha, 1))
    ia     = np.array(a).reshape((nalpha, 1))
    yi = ( ( (1. + ialpha) *
             np.abs(2. * (iX + idelta - np.trunc(iX + idelta)) - 1.)**ialpha +
             ia ) / (1. + ia) )
    y  = np.prod(yi, axis=0)

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def linear(X, a, b):
    """
    Linear test function

    .. math::
       Y = a*X + b

    Parameters
    ----------
    X : array_like
        (1,) or (1,npoints) array of floats
    a : array_like
        float or (npoints,) array of floats
    b : array_like
        float or (npoints,) array of floats

    Returns
    -------
    linear : float or ndarray
        float or (npoints,) floats of linear function values at *X* with
        parameters *a* and *b*

    """
    # Model output for given parameter set(s) is returned
    # X:    dim1 = # of parameters = 1
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] == 1, 'X.shape[0] must 1.'

    y = a * iX[0, :] + b

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def product(X):
    """
    Product test function

    .. math::
       Y = X[0] * X[1]

    Parameters
    ----------
    X : array_like
        (2,) or (2,npoints) array of floats

    Returns
    -------
    product : float or ndarray
        float or (npoints,) floats of product function values at *X*

    """
    # Model output for given parameter set(s) is returned
    # X:    dim1 = # of parameters = 2
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] == 2, 'X.shape[0] must 2.'

    y = iX[0, :] * iX[1, :]

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def ratio(X):
    """
    Ratio test function

    .. math::
       Y = X[0] / X[1]

    Simple nonlinear model proposed by Liu et al. (2006):

        Liu, H., Sudjianto, A., Chen, W., 2006.
        Relative entropy based method for probabilistic sensitivity analysis
        in engineering design.
        J. Mech. Des. 128, 326-336.

    Used by Pianosi & Wagener, Environmental Modelling & Software (2015)

        Pianosi, F. & Wagener T., 2015
        A simple and efficient method for global sensitivity analysis based on
        cumulative distribution functions.
        Environmental Modelling & Software 67, 1-11.

    Parameters
    ----------
    X : array_like
        (2,) or (2,npoints) array of floats

    Returns
    -------
    ratio : float or ndarray
        float or (npoints,) floats of ratio function values at *X*

    """
    # Model output for given parameter set(s) is returned
    # X:    dim1 = # of parameters = 2
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] == 2, 'X.shape[0] must 2.'

    y = iX[0, :] / iX[1, :]

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def ishigami_homma_easy(X):
    """
    Simplified Ishigami and Homma function

    .. math::
       Y = sin(X[0]) + X[1]

    with `X[0], X[1]` ~ `Uniform[-Pi, Pi]`

    Parameters
    ----------
    X : array_like
        (2,) or (2,npoints) array of floats

    Returns
    -------
    ishigami_homma_easy : float or ndarray
        float or (npoints,) floats of simplified Ishigami and Homma function
        values at *X*

    """
    # Model output for given parameter set(s) is returned
    # X:    dim1 = # of parameters = 2
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] == 2, 'X.shape[0] must 2.'

    y = np.sin(iX[0, :])  + iX[1, :]

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def ishigami_homma(X, a, b):
    """
    Ishigami and Homma (1990)

    given by Saltelli et al. (2008, page 179)

    Parameters
    ----------
    X : array_like
        (3,) or (3,npoints) array of floats
    a : array_like
        float or (npoints,) array of floats
    b : array_like
        float or (npoints,) array of floats

    Returns
    -------
    ishigami_homma : float or ndarray
        float or (npoints,) floats of Ishigami and Homma function values at
        *X* with parameters *a* and *b*

    """
    # Model output for given parameter set(s) is returned
    # X: dim1 = # of parameters = 3
    #       dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] == 3, 'X.shape[0] must be 3.'

    y = (np.sin(iX[0, :]) + a * (np.sin(iX[1, :]))**2 +
         b * iX[2, :]**4 * np.sin(iX[0, :]))

    if isone:
        return y[0]
    else:
        return y


# -----------------------------------------------------------


def K(X):
    """
    K example

    Saltelli et al. (2010) Comp. Phys. Comm., 181, p. 259-270

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats

    Returns
    -------
    K : float or ndarray
        float or (npoints,) floats of K function values at *X*

    """
    # Model output for given parameter set(s) is returned
    # X: dim1 = # of parameters = 10
    #    dim2 = # of parameter sets
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    assert iX.shape[0] > 1, 'X.shape[0] must be > 1.'

    nX = iX.shape[0]
    for ii in range(1, nX):
        iX[ii, :] = iX[ii - 1, :] * iX[ii, :]
    for ii in range(nX):
        iX[ii, :] = (-1)**(ii + 1) * iX[ii, :]
    y = iX.sum(axis=0)

    if isone:
        return y[0]
    else:
        return y


def bratley(*args):
    """
    K example

    Saltelli et al. (2010) Comp. Phys. Comm., 181, p. 259-270

    Parameters
    ----------
    X : array_like
        (nX,) or (nX,npoints) array of floats

    Returns
    -------
    bratley : float or ndarray
        float or (npoints,) floats of K function values at *X*

    """
    return K(*args)


# -----------------------------------------------------------


def fmorris(X, beta0, beta1, beta2, beta3, beta4):
    """
    Morris-function

    Morris (1991) Technometrics 33, 161-174

    Parameters
    ----------
    X : array_like
        (20,) or (20,npoints) array of floats
    beta0 : float
        float
    beta1 : array_like
        (20,) array of floats
    beta2 : array_like
        (20,20) array of floats
    beta3 : array_like
        (20,20,20) array of floats
    beta4 : array_like
        (20,20,20,20) array of floats

    Returns
    -------
    fmorris : float or ndarray
        float or (npoints,) floats of Morris function values at *X* with
        parameters *beta0-beta4*

    """
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    om = 2. * (iX - 0.5)
    ii = np.array((2, 4, 6))
    om[ii, :] = 2. * (1.1 * iX[ii, :] / (iX[ii, :] + 0.1) - 0.5)

    nn  = 20

    assert iX.shape[0] == nn, 'X.shape[0] must ' + str(nn) + '.'
    assert np.all(np.array(beta1.shape) == nn), (
        'beta1.shape must be all ' + str(nn) + '.')
    assert np.all(np.array(beta2.shape) == nn), (
        'beta2.shape must be all ' + str(nn) + '.')
    assert np.all(np.array(beta3.shape) == nn), (
        'beta3.shape must be all ' + str(nn) + '.')
    assert np.all(np.array(beta4.shape) == nn), (
        'beta4.shape must be all ' + str(nn) + '.')

    y    = np.empty(iX.shape[1])
    y[:] = beta0
    for i in range(0, nn):
        y[:] += beta1[i] * om[i, :]
        for j in range(i + 1, nn):
            y[:] += beta2[i, j] * om[i, :] * om[j, :]
            for k in range(j + 1, nn):
                y[:] += beta3[i, j, k] * om[i, :] * om[j, :] * om[k, :]
                for s in range(k + 1, nn):
                    y[:] += (beta4[i, j, k, s] *
                             om[i, :] * om[j, :] * om[k, :] * om[s, :])

    if isone:
        return y[0]
    else:
        return y


def morris(*args):
    """
    Morris-function

    Morris (1991) Technometrics 33, 161-174

    Parameters
    ----------
    X : array_like
        (20,) or (20, npoints) array of floats
    beta0 : float
        float
    beta1 : array_like
        (20,) array of floats
    beta2 : array_like
        (20, 20) array of floats
    beta3 : array_like
        (20, 20, 20) array of floats
    beta4 : array_like
        (20, 20, 20, 20) array of floats

    Returns
    -------
    morris : float or ndarray
        float or (npoints,) floats of Morris function values at *X* with
        parameters *beta0-beta4*

    """
    return fmorris(*args)


# -----------------------------------------------------------


def oakley_ohagan(X):
    """
    Oakley and O'Hagan (2004)

    J. R. Statist. Soc. B 66, Part 3, 751-769

    Parameters
    ----------
    X : array_like
        (15,) or (15, npoints) array of floats

    Returns
    -------
    oakley_ohagan : float or ndarray
        float or (npoints,) floats of Oakley and O'Hagan function values at *X*

    """
    X = np.array(X)
    if X.ndim == 1:
        isone = True
        iX = X[:, np.newaxis]
    else:
        isone = False
        iX = X

    nn = 15

    assert iX.shape[0] == nn, 'X.shape[0] must ' + str(nn) + '.'

    a1 = np.array([0.01, 0.05, 0.23, 0.04, 0.12, 0.39, 0.39, 0.61, 0.62, 0.40,
                   1.07, 1.15, 0.79, 1.12, 1.20])
    a2 = np.array([0.43, 0.09, 0.05, 0.32, 0.15, 1.04, 0.99, 0.97, 0.90, 0.81,
                   1.84, 2.47, 2.39, 2.00, 2.26])
    a3 = np.array([0.10, 0.21, 0.08, 0.27, 0.13, 0.75, 0.86, 1.03, 0.84, 0.80,
                   2.21, 2.04, 2.40, 2.05, 1.98])
    M  = np.array([
        [-0.02, -0.19, 0.13, 0.37, 0.17, 0.14, -0.44, -0.08, 0.71,
         -0.44, 0.5, -0.02, -0.05, 0.22, 0.06],
        [ 0.26, 0.05, 0.26, 0.24, -0.59, -0.08, -0.29, 0.42, 0.5,
          0.08, -0.11, 0.03, -0.14, -0.03, -0.22],
        [-0.06, 0.2, 0.1, -0.29, -0.14, 0.22, 0.15, 0.29, 0.23,
         -0.32, -0.29, -0.21, 0.43, 0.02, 0.04],
        [ 0.66, 0.43, 0.3, -0.16, -0.31, -0.39, 0.18, 0.06, 0.17,
          0.13, -0.35, 0.25, -0.02, 0.36, -0.33],
        [-0.12, 0.12, 0.11, 0.05, -0.22, 0.19, -0.07, 0.02, -0.1,
         0.19, 0.33, 0.31, -0.08, -0.25, 0.37],
        [-0.28, -0.33, -0.1, -0.22, -0.14, -0.14, -0.12, 0.22, -0.03,
         -0.52, 0.02, 0.04, 0.36, 0.31, 0.05],
        [-0.08, 0.004, 0.89, -0.27, -0.08, -0.04, -0.19, -0.36, -0.17,
         0.09, 0.4, -0.06, 0.14, 0.21, -0.01],
        [-0.09, 0.59, 0.03, -0.03, -0.24, -0.1, 0.03, 0.1, -0.34,
         0.01, -0.61, 0.08, 0.89, 0.14, 0.15],
        [-0.13, 0.53, 0.13, 0.05, 0.58, 0.37, 0.11, -0.29, -0.57,
         0.46, -0.09, 0.14, -0.39, -0.45, -0.15],
        [ 0.06, -0.32, 0.09, 0.07, -0.57, 0.53, 0.24, -0.01, 0.07,
          0.08, -0.13, 0.23, 0.14, -0.45, -0.56],
        [ 0.66, 0.35, 0.14, 0.52, -0.28, -0.16, -0.07, -0.2, 0.07,
          0.23, -0.04, -0.16, 0.22, 0, -0.09],
        [ 0.32, -0.03, 0.13, 0.13, 0.05, -0.17, 0.18, 0.06, -0.18,
          -0.31, -0.25, 0.03, -0.43, -0.62, -0.03],
        [-0.29, 0.03, 0.03, -0.12, 0.03, -0.34, -0.41, 0.05, -0.27,
         -0.03, 0.41, 0.27, 0.16, -0.19, 0.02],
        [-0.24, -0.44, 0.01, 0.25, 0.07, 0.25, 0.17, 0.01, 0.25,
         -0.15, -0.08, 0.37, -0.3, 0.11, -0.76],
        [ 0.04, -0.26, 0.46, -0.36, -0.95, -0.17, 0.003, 0.05, 0.23,
          0.38, 0.46, -0.19, 0.01, 0.17, 0.16] ])

    y = np.dot(a1, iX) + np.dot(a2, np.sin(iX)) + np.dot(a3, np.cos(iX))
    for i in range(iX.shape[1]):
        y[i] += np.dot(iX[:, i].T, np.dot(M, iX[:, i]))

    if isone:
        return y[0]
    else:
        return y.squeeze()


# -----------------------------------------------------------


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # ishigami_homma([np.pi/2.,np.pi/2.,1.], 1., 1.)
    # # 3.0
    # B(np.arange(10))
    # # 80
    # B(np.arange(12).reshape(6,2))
    # # array([56, 89])
    # linear(np.ones(1), 1., 1.)
    # # 2.0
    # product(np.arange(2)+1.)
    # # 2.0
    # ratio(np.arange(2)+1.)
    # # 0.5
    # ishigami_homma_easy([np.pi/2.,1.])
    # # 2.0
    # bratley(np.arange(5)+1.)
    # K(np.arange(5)+1.)
    # # -101.0
    # K(np.arange(8).reshape((4,2))+1.)
    # # [ 92., 342.]
    # oakley_ohagan(np.zeros(15))
    # # 15.75
    # Gstar(np.ones(5), np.zeros(5), np.ones(5), np.zeros(5))
    # # 1.0
    # g(np.ones(5), np.zeros(5))
    # G(np.ones(5), np.zeros(5))
    # # 32.0

    # # Morris function
    # seed = 1234
    # np.random.seed(seed=seed)
    # npars = 20
    # x0 = np.ones(npars)*0.5
    # lb = np.zeros(npars)
    # ub = np.ones(npars)
    # beta0              = 0.
    # beta1              = np.random.standard_normal(npars)
    # beta1[:10]         = 20.
    # beta2              = np.random.standard_normal((npars,npars))
    # beta2[:6,:6]       = -15.
    # beta3              = np.zeros((npars,npars,npars))
    # beta3[:5,:5,:5]    = -10.
    # beta4              = np.zeros((npars,npars,npars,npars))
    # beta4[:4,:4,:4,:4] = 5.
    # mm = fmorris(np.linspace(0,2*(npars-1),npars)/float(2*npars-1),
    #              beta0, beta1, beta2, beta3, beta4)
    # print(np.around(mm,3))
    # # -82.711
    # mm = fmorris(np.arange(2*npars,dtype=np.float).reshape((npars,2)) /
    #              float(2*npars-1),
    #              beta0, beta1, beta2, beta3, beta4)
    # print(np.around(mm,3))
    # # [-82.711 -60.589]



#!/usr/bin/env python
import numpy as np                       # array manipulation
import netCDF4 as nc
from pyjams.readnetcdf import readnetcdf


__all__ = ['writenetcdf', 'dumpnetcdf']


def writenetcdf(fhandle, vhandle=None, var=None, time=None, isdim=False,
                name=None, dims=None, attributes=None, fileattributes=None,
                comp=False, vartype=None, create_var=True):
    """Writes dimensions, variables, dependencies and attributes to NetCDF
    file for 1D to 6D data.

    All attributes must be lists; except the data itself.


    Definition
    ----------
    def writenetcdf(fhandle, vhandle=None, var=None, time=None, isdim=False,
                    name=None, dims=None, attributes=None, fileattributes=None,
                    comp=False, vartype=None, create_var=True):


    Input           Format                  Description
    -----           -----                   -----------
    fhandle         string                  file handle of nc.Dataset(FileName, 'w')
    vhandle         string                  variable handle of the particular variable
    var             array like              data (assumed to be netcdf float4)
    time            integer or array like   particular time step of the data
    isdim           boolean                 defines if current var is a dimension
    name            string                  name of the variable
    dims            1D list                 variable dependencies (e.g. [time, x, y])
    attributes      2D list or dictionary   variable attributes
    fileattributes  2D list or dictionary   global attributes of NetCDF file (history, description, ...)
    comp            boolean                 compress data on the fly using zlib
    vartype         string                  netcdf variable type
                                            default: 'f4' for normal variables
                                                     'f8' for variable with isdim=True and dims=None (=unlimited)
    create_var      boolean                 create variable for dimension although var is None

    Description
    -----------
    Open the NetCDF:
      fhandle =  nc.Dataset("Filename", 'w', format='NETCDF4')

    Then call writenetcdf with specifications of the dimensions:
      define
      - the dimension (isdim=True)
      - attributes of the dimensions (e.g. attributes=['units', 'm'])
      - the data of the dimensions as var=numpy.array or None (=unlimited dimension),

    Now write the data as variables:
    - when creating a variable, save the variable handle (vhandle)
    - specify the dimensions of data with dims
    - specify the timestep or timesteps with time


    Examples
    --------
    >>> import numpy as np
    >>> import netCDF4 as nc
    >>> dat = np.array([ [-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.],
    ...                  [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9.,-9.,-9., 5.],
    ...                  [ 5.,-9.,-9.,-9.,5., 5.,5.,-9., 5., 5., 5.],
    ...                  [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9., 5.,-9.,-9.],
    ...                  [-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.] ])

    # open file
    >>> fhandle = nc.Dataset('writenetcdf_test.nc', 'w', format='NETCDF4')

    # global attributes
    >>> FiAtt  = ([['description', 'test writing with writenetcdf.py'],
    ...            ['history', 'Created by Matthias Zink']])
    >>> handle = writenetcdf(fhandle, fileattributes=FiAtt)

    # dimension and dimension variables
    >>> varName = 'time'
    >>> dims    = None
    >>> varAtt  = ([['units', 'hours since 2011-01-01 00:00:00'],
    ...             ['calendar', 'gregorian']])
    >>> thand   = writenetcdf(fhandle, name=varName, dims=dims,
    ...                       attributes=varAtt, isdim=True)

    >>> varName = 'lon'
    >>> varAtt  = ([('units', 'degrees_east'),
    ...             ('standard_name', 'longitude'),
    ...             ('missing_value', -9.)])
    >>> dims    = dat.shape[0]
    >>> var     = np.arange(dat.shape[0])+1
    >>> handle  = writenetcdf(fhandle, name=varName, dims=dims, var=var,
    ...                       attributes=varAtt, isdim=True)

    >>> varName = 'lat'
    >>> varAtt  = ([['units', 'degrees_north'],
    ...             ['standard_name', 'latitude'],
    ...             ['missing_value', -9.]])
    >>> dims    = dat.shape[1]
    >>> var     = np.arange(dat.shape[1])+1
    >>> handle  = writenetcdf(fhandle, name=varName, dims=dims,
    ...                       var=var, attributes=varAtt, isdim=True)

    # times & variables
    >>> times  = np.array([0.5, 1., 1.5, 2.])
    >>> handle = writenetcdf(fhandle, thand, time=list(range(times.size)), var=times)

    >>> varName = 'TESTING'
    >>> varAtt  = {'units': 'm',
    ...            'long_name': 'Does this writing routine work?',
    ...            'missing_value': -9.}
    >>> dims    = ['time','lon','lat']
    >>> vhand   = writenetcdf(fhandle, name=varName, dims=dims,
    ...                       attributes=varAtt, comp=True)
    >>> for i in range(2):
    ...     handle = writenetcdf(fhandle, vhand, time=i, var=dat*(i+1))
    >>> handle = writenetcdf(fhandle, vhand, time=[2,3],
    ...                      var=np.array([dat*2, dat*3]))

    # type other than float
    >>> varName = 'TESTING2'
    >>> varAtt  = ([['units', 'm'],
    ...             ['long_name', 'Does this writing routine work?'],
    ...             ['_FillValue', -9]])
    >>> dims    = ['time','lon','lat']
    >>> typ     = 'i4'
    >>> vhand   = writenetcdf(fhandle, name=varName, dims=dims,
    ...                       attributes=varAtt, comp=True, vartype=typ)
    >>> for i in range(2):
    ...     handle = writenetcdf(fhandle, vhand, time=i,
    ...                          var=np.array(dat, dtype=int)*(i+1))

    # close file
    >>> fhandle.close()


    # check file
    >>> from readnetcdf import readnetcdf
    >>> print([ str(i) for i in
    ...         readnetcdf('writenetcdf_test.nc', variables=True) ])
    ['time', 'lon', 'lat', 'TESTING', 'TESTING2']
    >>> readdata = readnetcdf('writenetcdf_test.nc', var='TESTING')
    >>> print(np.any((readdata[0,:,:] - dat) != 0.))
    False
    >>> readdata2 = readnetcdf('writenetcdf_test.nc', var='TESTING2')
    >>> print(np.any((readdata2[0,:,:] - np.array(dat,dtype=int)) != 0.))
    False
    >>> if readdata.dtype == np.dtype('float32'): print('Toll')
    Toll
    >>> if readdata2.dtype == np.dtype('int32'): print('Toll')
    Toll

    >>> import os
    >>> os.remove('writenetcdf_test.nc')


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2020 Matthias Zink, Matthias Cuntz, Stephan Thober -
    mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Zink & Matthias Cuntz, Feb 2012
    Modified, Stephan Thober, May 2012 - type 'f8' for time dimension
              Matthias Cuntz, Jun 2012 - vartype
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Apr 2014 - attributes can be given
                  as dictionary, e.g. from readnetcdf with attributes=True
              Stephan Thober, May 2015 - added create_var flag that allows
                  to disable automatic creation of variables for dimensions
              Matthias Cuntz, Sep 2020 - _FillValue at variable creation
                                       - compatible with flake8
    """
    # create File attributes
    if fileattributes is not None:
        if type(fileattributes) is dict:
            for k in fileattributes:
                fhandle.setncattr(k, fileattributes[k])
        else:
            for i in range(len(fileattributes)):
                fhandle.setncattr(fileattributes[i][0], fileattributes[i][1])
        return None

    # get _FillValue from variable attributes if present
    fill_value = None
    if attributes is not None:
        if type(attributes) is dict:
            if '_FillValue' in attributes:
                fill_value = attributes.pop('_FillValue')
        else:
            anames  = [ aa[0] for aa in attributes ]
            avalues = [ aa[1] for aa in attributes ]
            if '_FillValue' in anames:
                fill_value = avalues[anames.index('_FillValue')]
                attributes = [ aa for aa in attributes
                               if aa[0] != '_FillValue' ]

    # create dimensions and dimension variables
    if vhandle is not None:
        hand = vhandle
    else:
        if not vartype:
            typ = 'f4'
            if isdim:
                if dims is not None:
                    typ = 'f8'
        else:
            typ = vartype
        if isdim:
            idim = fhandle.createDimension(name, dims)
            # create variable for the dimension
            if (var is not None) or create_var:
                hand = fhandle.createVariable(name, typ, (name,),
                                              fill_value=fill_value)
        else:
            keys = list(fhandle.dimensions.keys())
            for i in range(len(dims)):
                if dims[i] not in keys:
                    raise ValueError('Dimension ' + str(dims[i])
                                     + ' not in file dimensions: '
                                     + ''.join([i+' ' for i in keys]))
            hand = fhandle.createVariable(name, typ, tuple(dims), zlib=comp,
                                          fill_value=fill_value)

    if attributes is not None:
        if type(attributes) is dict:
            for k in attributes:
                hand.setncattr(k, attributes[k])
        else:
            for i in range(len(attributes)):
                hand.setncattr(attributes[i][0], attributes[i][1])

    if var is not None:
        shand = hand.shape
        if time is not None:
            svar = np.shape(var)
            if np.size(np.shape(time)) == 0:
                if np.size(svar) != (np.size(shand)-1):
                    raise ValueError('Variable and handle dimensions do not'
                                     + ' agree for variable time: '
                                     + str(svar) + ' and ' + str(shand))
            elif np.size(np.shape(time)) == 1:
                if np.size(svar) != np.size(shand):
                    raise ValueError('Variable and handle dimensions do not'
                                     + ' agree for variable time vector: '
                                     + str(svar) + ' and ' + str(shand))
            else:
                raise ValueError('Time must be scalar or index vector.')
            if np.size(shand) == 1:
                hand[time] = var
            elif np.size(shand) == 2:
                hand[time, :] = var
            elif np.size(shand) == 3:
                hand[time, :, :] = var
            elif np.size(shand) == 4:
                hand[time, :, :, :] = var
            elif np.size(shand) == 5:
                hand[time, :, :, :, :] = var
            elif np.size(shand) == 6:
                hand[time, :, :, :, :, :] = var
            else:
                raise ValueError('Number of dimensions not supported (>6): '
                                 + str(np.size(shand)))
        else:
            if np.size(var) != np.size(hand):
                raise ValueError('Variable and handle elements do not agree: '
                                 + str(np.size(var)) + ' and '
                                 + str(np.size(hand)))
            if np.size(shand) == 1:
                hand[:] = var
            elif np.size(shand) == 2:
                hand[:, :] = var
            elif np.size(shand) == 3:
                hand[:, :, :] = var
            elif np.size(shand) == 4:
                hand[:, :, :, :] = var
            elif np.size(shand) == 5:
                hand[:, :, :, :, :] = var
            elif np.size(shand) == 6:
                hand[:, :, :, :, :, :] = var
            else:
                raise ValueError('Number of dimensions not supported (>6): '
                                 + str(np.size(shand)))
    if (var is not None) or create_var:
        return hand
    else:
        return


# returns list of all dimensions given a filename
def _get_dims(fname):
    """
    This functions returns all dimension names contained
    in a netcdf file as list.


    Definition
    ----------
    def _get_dims(fname)


    Input           Format                  Description
    -----           -----                   -----------
    fname           string                  filename of netcdf file


    Description
    -----------
    reads all dimension names from netcdf file
    dnames = _get_dims('test.nc')


    Examples
    --------
    >>> import numpy as np
    >>> dat = np.array([[-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9.,-9.,-9., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5., 5.,5.,-9., 5., 5., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9., 5.,-9.,-9.],
    ...                 [-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.]])
    >>> dims  = ['xx', 'yy']
    >>> FiAtt = ([['description', 'test dump_netcdf'],
    ...           ['history', 'Created by Stephan Thober']])
    >>> dumpnetcdf('test_dump.nc', dims=dims, fileattributes=FiAtt,
    ...            data=(dat, {'long_name':'CHS'}))

    # check file
    >>> import sys
    >>> pyver = sys.version_info
    >>> from readnetcdf import readnetcdf
    >>> out = _get_dims('test_dump.nc')
    >>> print(out) if pyver > (3,0) else print([ i.encode('UTF-8')
    ...                                          for i in out ])
    ['xx', 'yy']

    >>> import os
    >>> os.remove('test_dump.nc')


    History
    -------
    Written,  Stephan Thober, Jun 2015
    Modified, Matthias Cuntz, Nov 2016 - adapted docstring to Python 2 and 3
              Matthias Cuntz, Sep 2020 - compatible with flake8
    """
    file_vars = readnetcdf(fname, var='', variables=True)
    file_dims = []
    for var in file_vars:
        tmp_dims = readnetcdf(fname, var=var, dims=True)
        for dims in tmp_dims:
            if dims not in file_dims:
                file_dims.append(dims)
    return file_dims


# write to file
def dumpnetcdf(fname, dims=None, fileattributes=None, vnames=None,
               create=True, **variables):
    """
    Writes variables with the same dimension to a netcdf file (1D-5D).
    It is a wrapper around writenetcdf.

    Variables must be given as keyword arguments. The key becomes the name of
    the variable in the netcdf file. The value becomes the array.


    Definition
    ----------
    def dumpnetcdf(fname, dims=None, fileattributes=None, vnames=None,
                   create=True, **variables):


    Input           Format                  Description
    -----           -----                   -----------
    fname           string                  filename of netcdf file
    dims            1D list of strings      dimension names (e.g., ['time', 'x', 'y']),
                                            default is [ 'x', 'y', 'z', 'u', 'v' ]
    fileattributes  2D list or dictionary   global attributes of NetCDF file
                                            (e.g., history, description, ...)
    create          bool                    flag for creating file or appending variables
    vnames          dictionary              dictionary that specifies the variable name
                                            for each given variable
    variables       keyword arguments       keys become variable names, values are either
                                            numpy arrays to write OR list/tuple of
                                            [ numpy array, attributes ] where attributes
                                            must be a dictionary

    Description
    -----------
    write netcdf file in one single line
    dump_netcdf('test.nc', dims=[ 'time', 'northing', 'easting' ],
                 sm=sm_array,
                 smi=[ smi_array, {'long_name': 'soil moisture index'} ])

    Restrictions
    ------------
    all arrays must have the same dimensions and must be numpy arrays

    Examples
    --------
    >>> import numpy as np
    >>> dat = np.array([[-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9.,-9.,-9., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5., 5.,5.,-9., 5., 5., 5.],
    ...                 [ 5.,-9.,-9.,-9.,5.,-9.,5.,-9., 5.,-9.,-9.],
    ...                 [-9., 5., 5.,-9.,5.,-9.,5.,-9., 5., 5., 5.]])
    >>> dims  = ['xx', 'yy']
    >>> FiAtt = ([['description', 'test dump_netcdf'],
    ...           ['history', 'Created by Stephan Thober']])
    >>> dumpnetcdf('test_dump.nc', dims=dims, fileattributes=FiAtt,
    ...            data=(dat, {'long_name':'CHS'}))

    # check file
    >>> from readnetcdf import readnetcdf
    >>> print([str(i) for i in readnetcdf('test_dump.nc', variables=True)])
    ['data']
    >>> readdata = readnetcdf('test_dump.nc', var='data')
    >>> print(np.any((readdata[:,:] - dat) != 0.))
    False
    >>> if readdata.dtype == np.dtype('float64'): print('Toll')
    Toll

    >>> import os
    >>> os.remove('test_dump.nc')


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2016 Stephan Thober

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Stephan Thober, Sep 2014
    Modified, Stephan Thober, Jan 2015 - bug fix, only parse ndim dimensions
                  to writenetcdf
              Stephan Thober & Matthias Zink, Mar 2015 - added flag to append
                  variables
              Stephan Thober, Jun 2016 - included dump for single numbers
                  without attributes
              Stephan Thober, Aug 2016 - included separate variable names
              Matthias Cuntz, Nov 2016 - ported to Python 3, mostly dictionary
                  behaviour
              Matthias Cuntz, Sep 2020 - compatible with flake8
    """
    # check that variables are given
    if variables == dict():
        raise ValueError('Variables must be given as keyword arguments.')

    # check if dims are given
    if dims is None:
        dims = [ 'x', 'y', 'z', 'u', 'v' ]

    # check if variable names are given
    if vnames is not None:
        if type(vnames) != dict:
            raise ValueError('Variable names must be dictionary with the'
                             + 'same keys as given variables')
        if len(vnames) != len(variables):
            raise ValueError('Number of given variable names does not'
                             + 'match number of variables')
    # get dimensions from first variable
    vv1 = variables[list(variables.keys())[0]]
    if isinstance(vv1, (tuple, list)):
        arr_shape = vv1[0].shape
    else:
        arr_shape = vv1.shape
    # open netcdf file
    if create:
        fh = nc.Dataset(fname, 'w', 'NETCDF4')
    else:
        fh = nc.Dataset(fname, 'a', 'NETCDF4')
    # write file attributes
    if fileattributes is not None:
        writenetcdf(fh, fileattributes=fileattributes)
    # create dimensions according to first variable
    if create:
        for dd in range(len(arr_shape)):
            writenetcdf(fh, name=dims[dd], dims=arr_shape[dd],
                        var=None, isdim=True, create_var=False)
    else:
        # read dimensions from file
        file_dims = _get_dims(fname)
        for dd in range(len(arr_shape)):
            # write dimensions if they do not exist
            if not dims[dd] in file_dims:
                writenetcdf(fh, name=dims[dd], dims=arr_shape[dd],
                            var=None, isdim=True, create_var=False)
    # loop over variables
    cc = 0
    for key in variables:
        value = variables[key]
        # update vnames if required
        if vnames is not None:
            key = vnames[key]
            cc += 1
        if len(value) == 1:
            # WRITE SINGLE NUMBER WITHOUT ATTRIBUTE
            if value.ndim > len(dims):
                raise ValueError('***size mismatch: variable '+key)
            writenetcdf(fh, name=key, dims=dims[:value.ndim], var=value,
                        vartype=value.dtype, comp=True)
        else:
            if type( value[-1] ) == dict:
                # WRITE WITH ATTRIBUTE
                if value[0].ndim > len(dims):
                    raise ValueError('***size mismatch: variable '+key)
                writenetcdf(fh, name=key, dims=dims[:value[0].ndim],
                            var=value[0], vartype=value[0].dtype,
                            attributes=value[-1], comp=True)
            else:
                # WRITE ARRAY WITHOUT ATTRIBUTE
                if value.ndim > len(dims):
                    raise ValueError('***size mismatch: variable '+key)
                writenetcdf(fh, name=key, dims=dims[:value.ndim], var=value,
                            vartype=value.dtype, comp=True)
    fh.close()


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
import pyjams
from collections import OrderedDict


__all__ = ['read_data', 'write_data', 'write_data_dmp', 'write_data_dmp_size','write_data_one_file']


# --------------------------------------------------------------------

def read_data(files, undef=-9999., strip=None, norecord=False, nofill=False):
    """
        Read and concatenate data from CHS level1 data files.


        Definition
        ----------
        def read_data(files, undef=-9999., strip=None, norecord=False, nofill=False):


        Input
        -----
        files     (nfile,)-list with CHS data level1 file names


        Optional Input
        --------------
        undef     fill value for data and flags if non-existant (default: -9999.)
        strip     Strip strings with str.strip(strip) during read.
                  If None then strip quotes " and ' (default).
                  If False then no strip (30% faster).
                  Otherwise strip character given by strip.
        norecord  True: Do not assume that second column is record number.
        nofill    True: do not fill-up equal distant time steps


        Output
        ------
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead
            where
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        if norecord==False: record    (n,)-array of record number
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        if norecord==False: hrecord   record header
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data(files, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead)


        # Read data without record
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True, nofill=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data(files, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead)

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2017 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015
        Modified, MC, May 2015 - different variable in different input files
                               - strip
                               - norecord
                               - nofill
                  MC, Nov 2017 - assert files is iterable except string
    """

    debug = False
    iundef = np.int(undef)
    if not isinstance(files, (list, tuple, np.ndarray)): files = [files]

    # Get unique header and time stamps of all input file
    for cff, ff in enumerate(files):
        ihead = pyjams.fread(ff, skip=1, strip=strip, header=True)        # head with TIMESTAMP and RECORD
        if not norecord:
            if (ihead[1].split()[0]).lower() != 'record':
                raise ValueError('Assume the following structure: Date, Record, data, flag, data, flag, ...')
        idate = pyjams.sread(ff, skip=1, nc=1, squeeze=True, strip=strip) # TIMESTAMP

        idate = pyjams.date2en(idate, full=True)
        if cff == 0:
            if norecord:
                hdat   = ihead[1::2]
                hflags = ihead[2::2]
            else:
                hdat   = ihead[2::2]
                hflags = ihead[3::2]
            adate  = idate
        else:
            if norecord:
                hdat   += ihead[1::2]
                hflags += ihead[2::2]
            else:
                hdat   += ihead[2::2]
                hflags += ihead[3::2]
            adate  += idate
    hdat   = list(OrderedDict.fromkeys(hdat))   # unique data head
    hflags = list(OrderedDict.fromkeys(hflags)) # unique flags head
    adate  = list(OrderedDict.fromkeys(adate))  # unique dates
    ii     = pyjams.argsort(hdat)
    hdat   = [ hdat[i] for i in ii ]
    hflags = [ hflags[i] for i in ii ]
    adate.sort()

    # Fill missing time steps in all time steps
    if not nofill:
        date  = pyjams.jams.date2dec(eng=adate)
        dd    = np.round(np.diff(date)*24.*60.).astype(np.int) # minutes between time steps
        dmin  = np.amin(dd)                                    # time step in minutes
        dt    = np.float(dmin)/(24.*60.)                       # time step in fractional day
        igaps = np.where(dd != dmin)[0]                        # indexes of gaps
        for i in igaps[::-1]:
            nt      = np.round((date[i+1]-date[i])/dt).astype(np.int) # # of missing dates
            newdate = (date[i]+np.arange(1,nt)*dt)[::-1]              # the missing dates in reverse order
            for j in range(nt-1):
                date.insert(i+1, newdate[j]) # fill in missing dates, last one first
        adate = pyjams.jams.dec2date(date, eng=True)

    # Read files and fill in output array
    nrow   = len(adate)
    ncol   = len(hdat)
    dat    = np.ones((nrow,ncol))*undef # output array without
    flags  = np.ones((nrow,ncol), dtype=np.int)*iundef           # output array
    if not norecord: record = np.ones(nrow, dtype=np.int)*iundef # output array
    iidate = list()                     # list with indices of dates in dat/flag/record arrays
    iihead = list()                     # list with indices of header in hdat/hflags lists
    for cff, ff in enumerate(files):
        if debug: print('File name: ', ff)
        # date, data
        isdat, ssdat = pyjams.fsread(ff, skip=1, snc=[0], nc=-1, strip=strip, strarr=True) # array
        isdate = ssdat[:,0]
        if norecord:
            idat    = isdat[:,0::2]
            iflags  = isdat[:,1::2].astype(np.int)
        else:
            irecord = isdat[:,0].astype(np.int)
            idat    = isdat[:,1::2]
            iflags  = isdat[:,2::2].astype(np.int)
        isdate  = pyjams.date2en(isdate, full=True)
        # date header, data header
        ihead, shead = pyjams.fsread(ff, skip=1, snc=[0], nc=-1, strip=strip, header=True) # list
        ihdate   = shead[0]
        if norecord:
            ihdat    = ihead[0::2]
            ihflags  = ihead[1::2]
        else:
            ihrecord = ihead[0]
            ihdat    = ihead[1::2]
            ihflags  = ihead[2::2]
        # date and record header
        if cff == 0:
            hdate   = ihdate
            if not norecord: hrecord = ihrecord
        else:
            if norecord:
                if (ihdate != hdate):
                    raise ValueError('Assume the same date headers.')
            else:
                if (ihdate != hdate) or (ihrecord != hrecord):
                    raise ValueError('Assume the same date and record headers.')

        # fill output arrays and index lists
        iiidate = np.where(np.in1d(adate,isdate))[0]  # indexes in dat/flags of time steps in current file
        if debug:
            def findDuplicates(l):      # this function returns all duplicates in a list
                return list([x for x in l if l.count(x) > 1])

            print('Numbers should match: ', iiidate.size, idat.shape[0])
            print('duplicate dates: ',findDuplicates(list(isdate)))
            # print('Current time steps:  ', len(isdate), [ aa for aa in isdate ])
            # print('Selected time steps: ', len(iiidate), [ adate[aa] for aa in iiidate ])                

        iidate.append(iiidate)
        if not norecord: record[iiidate] = irecord[:] # write at appropriate places in record
        iiihead = list()
        for i, h in enumerate(ihdat):
            hh = hdat.index(h)
            iiihead.extend([hh])
            dat[iiidate, hh]   = idat[:,i]            # write at appropriate places in dat
            flags[iiidate, hh] = iflags[:,i]          # write at appropriate places in flags
        iihead.append(iiihead)

    if norecord:
        return adate, dat, flags, iidate, hdate, hdat, hflags, iihead
    else:
        return adate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead

# --------------------------------------------------------------------

def write_data(*args):
    """
        Write concatenated data back to individual CHS level1 data files.

        Wrapper to write_data_norecord and write_data_record.


        Definition
        ----------
        def write_data(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead):
        or
        def write_data(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        [record   (n,)-array of record number]
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        [hrecord  record header]
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data(files, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead)


        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data(files, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead)

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2015
    """
    if len(args) == 9:
        write_data_norecord(*args)
    elif len(args) == 11:
        write_data_record(*args)
    else:
        raise ValueError('Must have 9 or 11 arguments.')

# --------------------------------------------------------------------

def write_data_dmp(*args):
    """
        Write data to individual Tereno Level2b files.

        Wrapper to write_data_norecord_dmp.


        Definition
        ----------
        def write_data_dmp(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead, hdmp):
        or
        def write_data_dmp(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        [record   (n,)-array of record number]
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        [hrecord  record header]
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files
        hdmp      data headers in Data Management Portal (DMP)


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout (DB)')
        jams.level1.write_data_dmp(ofiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead, hdmp)


        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout (DB)')
        jams.level1.write_data_dmp(ofiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp)

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2015
    """
    if len(args) == 10:
        write_data_norecord_dmp(*args)
    elif len(args) == 12:
        write_data_norecord_dmp(args[0],args[1],args[3],args[4],args[5],args[6],args[8],args[9],args[10],args[11])
    else:
        raise ValueError('Must have 10 or 12 arguments.')


# --------------------------------------------------------------------

def write_data_dmp_size(*args):
    """
        Write data to individual Tereno Level2b files and split if file size exceeds 10 MB.

        Wrapper to write_data_norecord_dmp_size.


        Definition
        ----------
        def write_data_dmp(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead, hdmp):
        or
        def write_data_dmp(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        [record   (n,)-array of record number]
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        [hrecord  record header]
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files
        hdmp      data headers in Data Management Portal (DMP)


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout (DB)')
        jams.level1.write_data_dmp(ofiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead, hdmp)


        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout (DB)')
        jams.level1.write_data_dmp(ofiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp)

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2016 Matthias Cuntz, Benjamin Dechant - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2015
        Modified, BD, June 2016
    """
    if len(args) == 10:
        write_data_norecord_dmp_size(*args)
    elif len(args) == 12:
        write_data_norecord_dmp_size(args[0],args[1],args[3],args[4],args[5],args[6],args[8],args[9],args[10],args[11])
    else:
        raise ValueError('Must have 10 or 12 arguments.')


# --------------------------------------------------------------------

def write_data_one_file(*args):
    """
        Write concatenated data back to one CHS data file.

        Wrapper to write_data_norecord_one_file and write_data_record_one_file.


        Definition
        ----------
        def write_data_one_file(infiles, sdate, record, dat, flags, hdate, hrecord, hdat, hflags):
        or
        def write_data_one_file(infiles, sdate, dat, flags, hdate, hdat, hflags):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        [record   (n,)-array of record number]
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        hdate     date/time header
        [hrecord  record header]
        hdat      data headers
        hflags    flags headers


        Output
        ------
        file will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofile = files[0].replace('level1', 'level2')
        jams.level1.write_data_one_file(ofile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags)


        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofile = files[0].replace('level1', 'level2')
        jams.level1.write_data_one_file(ofile, sdate, dat, flags, hdate, hdat, hflags)

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, May 2016
    """
    if len(args) == 7:
        write_data_norecord_one_file(*args)
    elif len(args) == 9:
        write_data_record_one_file(*args)
    else:
        raise ValueError('Must have 7 or 9 arguments.')

# --------------------------------------------------------------------

def write_data_norecord(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead):
    """
        Write concatenated data back to individual CHS level1 data files without record column.


        Definition
        ----------
        def write_data_norecord(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data_norecord(files, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015 - from write_data
    """
    # Assure iterable infiles
    if not isinstance(infiles, (list, tuple, np.ndarray)): infiles = [infiles]

    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(infiles)   == len(iidate),  'File list and date index list do not conform.'
    assert len(infiles)   == len(iihead),  'File list and header index list do not conform.'
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'

    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Write individual files
    for cff, ff in enumerate(infiles):
        f = open(ff, 'w')
        # header
        hstr = hdate
        ihead = iihead[cff]
        for i in ihead:
            hstr += ','+hdat[i]+','+hflags[i]
        print(hstr, file=f)
        # data
        idate = iidate[cff]
        for j in idate:
            dstr = isdate[j]
            for i in ihead:
                # dstr += ','+str(dat[j,i])+',-9999'   # for test: restate -9999 for all flags:
                dstr += ','+str(dat[j,i])+','+str(flags[j,i])
            print(dstr, file=f)
        f.close()

# --------------------------------------------------------------------

def write_data_norecord_dmp(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):
    """
        Write data to individual Tereno Level2b files.


        Definition
        ----------
        def write_data_norecord_dmp(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):


        Input
        -----
        infiles   (nfile,)-list with CHS data level2b file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files
        hdmp      data headers in Data Management Portal (DMP)


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout(DB)')
        jams.level1.write_data_norecord_dmp(ofiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2016 - from write_data_norecord
    """
    # Assure iterable infiles
    if not isinstance(infiles, (list, tuple, np.ndarray)): infiles = [infiles]

    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(infiles)   == len(iidate),  'File list and date index list do not conform.'
    assert len(infiles)   == len(iihead),  'File list and header index list do not conform.'
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'
    assert len(hdmp)      == ncol,         'Not enough database headers.'

    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Tereno flags: OK, DOUBTFUL or BAD
    if np.any(flags > 2): # level1 flags
        oflags = np.maximum(pyjams.jams.level1.get_maxflag(flags), 0)
    else:                 # level2 flags
        oflags = flags
    strflags = np.zeros(oflags.shape, dtype='S'+str(len('DOUBTFUL,Other,From CHS flagging'))) #NIL
    ii = np.where(oflags == 0)
    if ii[0].size>0: strflags[ii] = 'OK,NIL,NIL'
    ii = np.where(oflags == 1)
    if ii[0].size>0: strflags[ii] = 'DOUBTFUL,Other,From CHS flagging'
    ii = np.where(oflags == 2)
    if ii[0].size>0: strflags[ii] = 'BAD,Other,From CHS flagging'
   
    

    # Write individual files
    for cff, ff in enumerate(infiles):
        f = open(ff, 'w')
        # header
        hstr = 'timestamp,sensorname,value,quality_flag,quality_cause,quality_comment'
        print(hstr, file=f)
        # data
        ihead = iihead[cff]
        idate = iidate[cff]
        for j in idate:
            for i in ihead:
                dstr = isdate[j]+','+hdmp[i]+','+str(dat[j,i])+','+strflags[j,i]
                print(dstr, file=f)
        f.close()

# --------------------------------------------------------------------

def write_data_norecord_dmp_size(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):
    """
        Write data to individual Tereno Level2b files. If the files sizes exceeds 10 MB, the level2b file is splitted into chunks < 10 MB


        Definition
        ----------
        def write_data_norecord_dmp(infiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp):


        Input
        -----
        infiles   (nfile,)-list with CHS data level2b file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files
        hdmp      data headers in Data Management Portal (DMP)


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofiles = [ f.replace('level2','level2b') for f in files ]
        hdmp = jams.level1.get_value_excel(chsxlsfile, sheet, hdat, 'headerout(DB)')
        jams.level1.write_data_norecord_dmp(ofiles, sdate, dat, flags, iidate, hdate, hdat, hflags, iihead, hdmp)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2016 - from write_data_norecord
        Modified, MD, June 2016 file size limit
    """
    # Assure iterable infiles
    if not isinstance(infiles, (list, tuple, np.ndarray)): infiles = [infiles]

    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(infiles)   == len(iidate),  'File list and date index list do not conform.'
    assert len(infiles)   == len(iihead),  'File list and header index list do not conform.'
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'
    assert len(hdmp)      == ncol,         'Not enough database headers.'


    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Tereno flags: OK, DOUBTFUL or BAD
    if np.any(flags > 2): # level1 flags
        oflags = np.maximum(pyjams.jams.level1.get_maxflag(flags), 0)
    else:                 # level2 flags
        oflags = flags
    strflags = np.zeros(oflags.shape, dtype='S'+str(len('DOUBTFUL,Other,From CHS flagging'))) #NIL
    ii = np.where(oflags == 0)
    if ii[0].size>0: strflags[ii] = 'OK,NIL,NIL'
    ii = np.where(oflags == 1)
    if ii[0].size>0: strflags[ii] = 'DOUBTFUL,Other,From CHS flagging'
    ii = np.where(oflags == 2)
    if ii[0].size>0: strflags[ii] = 'BAD,Other,From CHS flagging'
    
    

    # Write individual files and split if size > 10 MB
    for cff, ff in enumerate(infiles):
    
        k = 0
        start_index = 0

        # data
        ihead = iihead[cff]
        idate = iidate[cff]

        while start_index < iidate[0][-1]:
            
            if start_index == 0:
                f = open(ff, 'w')
            else:
                f = open(ff[0:-12]+'_'+str(k)+ff[-12:], 'w')

            k += 1
            # header
            hstr = 'timestamp,sensorname,value,quality_flag,quality_cause,quality_comment'
            print(hstr, file=f)
            Line = []
            for i in ihead:
                Line.append(isdate[0]+','+hdmp[0]+','+str(dat[0,i])+','+strflags[0,i])
    
            Len = len(str(Line))
            numLines = 0
            for j in idate[start_index:]:
                for i in ihead:
                    numLines +=1
                    dstr = isdate[j]+','+hdmp[i]+','+str(dat[j,i])+','+strflags[j,i]
                    print(dstr, file=f)
        
                f.seek(0,2)
                size = f.tell()
                if(size > 9700000-1*Len or numLines > 100000 - len(ihead)-1):    # limit to less than 10 MB or 100 000 lines
                    print('reached max. size')
                    start_index = j
                    break
                else:
                    start_index = iidate[0][-1]
               
            f.close()
        
# --------------------------------------------------------------------

def write_data_norecord_one_file(infile, sdate, dat, flags, hdate, hdat, hflags):
    """
        Write concatenated data back to one CHS data file without record column.


        Definition
        ----------
        def write_data_norecord_one_file(infile, sdate, dat, flags, hdate, hdat, hflags):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        hdate     date/time header
        hdat      data headers
        hflags    flags headers


        Output
        ------
        file will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, dat, flags, iidate, hdate, hdat, hflags, iihead = jams.level1.read_data(files, norecord=True)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        ofile = files[0].replace('level1', 'level2')
        jams.level1.write_data_norecord_one_file(ofile, sdate, dat, flags, hdate, hdat, hflags)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2016 - from write_data_norecord
    """
    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'

    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Write individual files
    f = open(infile, 'w')
    # header
    hstr = hdate
    for i in range(ncol):
        hstr += ','+hdat[i]+','+hflags[i]
    print(hstr, file=f)
    # data
    for j in range(ntime):
        dstr = isdate[j]
        for i in range(ncol):
            dstr += ','+str(dat[j,i])+','+str(flags[j,i])
        print(dstr, file=f)
    f.close()

# --------------------------------------------------------------------

def write_data_record(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead):
    """
        Write concatenated data back to individual CHS level1 data files with record number.


        Definition
        ----------
        def write_data_record(infiles, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        record    (n,)-array of record number
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        iidate    (nfile,)-list with indices in the output arrays of the input files
        hdate     date/time header
        hrecord   record header
        hdat      data headers
        hflags    flags headers
        iihead    (nfile,)-list with indices in the output array of headers in the input files


        Output
        ------
        files will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data
        jams.level1.write_data_record(files, sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015
        Modified, MC, May 2015 - different variable in different input files
                               - renamed to write_data_record
    """
    # Assure iterable infiles
    if not isinstance(infiles, (list, tuple, np.ndarray)): infiles = [infiles]

    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(infiles)   == len(iidate),  'File list and date index list do not conform.'
    assert len(infiles)   == len(iihead),  'File list and header index list do not conform.'
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert len(record)    == ntime,        'Not enough record numbers.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'

    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Write individual files
    for cff, ff in enumerate(infiles):
        f = open(ff, 'w')
        # header
        hstr = hdate+','+hrecord
        ihead = iihead[cff]
        for i in ihead:
            hstr += ','+hdat[i]+','+hflags[i]
        print(hstr, file=f)
        # data
        idate = iidate[cff]
        for j in idate:
            dstr = isdate[j]+','+str(record[j])
            for i in ihead:
                # dstr += ','+str(dat[j,i])+',-9999'   # for test: restate -9999 for all flags:
                dstr += ','+str(dat[j,i])+','+str(flags[j,i])
            print(dstr, file=f)
        f.close()

# --------------------------------------------------------------------

def write_data_record_one_file(infile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags):
    """
        Write concatenated data back to one CHS data file with record number.


        Definition
        ----------
        def write_data_record_one_file(infile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags):


        Input
        -----
        infiles   (nfile,)-list with CHS data level1 file names
        sdate     (n,)-array of ascii dates in format YYYY-MM-DD hh:mm:ss
        record    (n,)-array of record number
        dat       (n,m)-array of data
        flags     (n,m)-array of flags
        hdate     date/time header
        hrecord   record header
        hdat      data headers
        hflags    flags headers


        Output
        ------
        file will be overwritten


        Examples
        --------
        --> see __init__.py for full example of workflow

        # Read data
        files = jams.files_from_gui(title='Choose Level 1 file(s)')
        sdate, record, dat, flags, iidate, hdate, hrecord, hdat, hflags, iihead = jams.level1.read_data(files)

        # Set flags if variables were not treated yet
        flags[:,idx] = np.where(flags[:,idx]==np.int(undef), 9, flags[:,idx])

        # Write back data to one file
        ofile = files[0].replace('level1', 'level2')
        jams.level1.write_data_record_one_file(ofile, sdate, record, dat, flags, hdate, hrecord, hdat, hflags)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2016 - from write_data_record
    """
    # Few checks of sizes
    ntime = dat.shape[0]
    ncol  = dat.shape[1]
    assert len(sdate)     == ntime,        'Not enough dates.'
    assert len(record)    == ntime,        'Not enough record numbers.'
    assert flags.shape[0] == ntime,        'Not enough flag time steps.'
    assert flags.shape[1] == ncol,         'Not enough flag columns.'
    assert len(hdat)      == ncol,         'Not enough data headers.'
    assert len(hflags)    == ncol,         'Not enough flag headers.'

    # assure YYYY-MM-DD hh:mm:ss format even if sdate has DD.MM.YYYY hh:m:ss format
    isdate = pyjams.date2en(sdate, full=True)

    # Write individual files
    f = open(infile, 'w')
    # header
    hstr = hdate+','+hrecord
    for i in range(ncol):
        hstr += ','+hdat[i]+','+hflags[i]
    print(hstr, file=f)
    # data
    for j in range(ntime):
        dstr = isdate[j]+','+str(record[j])
        for i in range(ncol):
            dstr += ','+str(dat[j,i])+','+str(flags[j,i])
        print(dstr, file=f)
    f.close()

# --------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
